use salvo::{handler, hyper::header::HOST, prelude::StatusCode, Request, Response};
use tracing::{debug, error, info};

use crate::{
    db,
    handlers::{
        convert_file_size, convert_unix_timestamp, count_file_metrics, render_template,
        TemplateStruct,
    },
    SQLITE,
};

use super::guess_ip;

/// This file contains the handler that serves the users stats back to them.
/// We expect a API Key to be provided in the cookie or header here.
/// This will list only files that were uploaded with the provided API Key.
#[handler]
pub async fn list_files(req: &mut Request, res: &mut Response) {
    // Setup db pool
    let sqlconn = SQLITE.get().unwrap();

    // Get the IP of the person viewing, so we can show their files
    let headers = req.headers();
    let remote_addr = &req.remote_addr().clone();
    let ip = guess_ip(headers, remote_addr);

    debug!("list_files(remote_addr, ip): {:?}, {:?}", remote_addr, ip);

    // This returns a Vec of all the files uploaded by a specific IP Address
    let files = db::get_my_files(sqlconn, &ip).await;

    let mut html = String::new();
    // For each file in the list, add the html to the rendered string
    match files.as_ref() {
        Ok(file_metrics) => {
            for f in file_metrics {
                // Change the filename to a file url
                let fileurl = format!("/{}", &f.filename);

                // Change the filesize to a human readable value
                let human_size = convert_file_size(f.filesize);

                // Change the expiry to a human readable value
                let human_time = convert_unix_timestamp(f.expiry);

                // If isDeleted, change colour to red, else green
                let colour = if f.is_deleted {
                    "background-color: #ff00000a;".to_string()
                } else {
                    "background-color: #00ff100a;".to_string()
                };

                // Add a new file to the rendered string
                html.push_str(&format!("<tr style='{}'><td><a href={}>{}</a></td><td data-sort='{}'>{}</td><td>{}</td><td>{}</td><td data-sort='{}'>{}</td></tr>",
                    colour,
                    fileurl.as_str(),
                    f.filename.as_str(),
                    f.filesize,
                    human_size.as_str(),
                    f.mimetype.as_str(),
                    f.views.to_owned(),
                    f.expiry.to_owned(),
                    human_time.to_owned(),
                ));
            }
        }
        Err(err) => {
            // Log error to console
            error!("Error retrieving file metrics: {:?}", err);

            // Render a nice error page for the poor user
            let template_filename = "error.html";
            let template = TemplateStruct {
                domain: String::from(headers[HOST].to_str().unwrap()),
                message1: String::from("Error 500: Internal Server Error"),
                message2: String::from(
                    "Something went wrong while reading the result from the database, uh oh :(",
                ),
                ..Default::default()
            };

            let status_code = StatusCode::INTERNAL_SERVER_ERROR;
            render_template(res, headers, template_filename, template, status_code).await;
        }
    }

    // Close the table when all files are 'rendered'
    html.push_str("</table>");

    // Log this to the server console
    info!(
        "[{}]New Request: /my_files ({} Files)",
        &headers[HOST].to_str().unwrap(),
        count_file_metrics(&files)
    );

    let template_filename = "my_files.html";
    let template = TemplateStruct {
        domain: String::from(headers[HOST].to_str().unwrap()),
        message1: html,
        ..Default::default()
    };
    let status_code = StatusCode::OK;

    render_template(res, headers, template_filename, template, status_code).await
}
