use figment::providers::{Format, Toml};
use figment::Figment;
use magic::Cookie;
use once_cell::sync::OnceCell;
use salvo::prelude::*;
use salvo::serve_static::{StaticDir, StaticFile};
use sqlx::SqlitePool;

use lazy_static::lazy_static;
use std::fs::create_dir_all;
use tracing_subscriber::filter::EnvFilter;
use tracing_subscriber::fmt;
use tracing_subscriber::prelude::*;

use crate::config::AppConfig;
use crate::db::DatabaseType;
use crate::handlers::cleaner_thread;

// Import sub-modules.
mod config;
mod db;
mod engine;

// Import tests for main, db and engine.
mod db_tests;
#[cfg(test)]
mod engine_tests;

// Import the web request handlers modules.
mod handlers;

// I /don't think/ there's any harm to adding multiple of thse for each sql backend.
// Setup the global sqlite db
static SQLITE: OnceCell<SqlitePool> = OnceCell::new();

// Initialise the magic db
// This is thread-local, so each thread will have it's own database loaded.
// It's probably not the best way of doing it
// and we could probably have a dedicated thread for mimetypes but idk i'm not that smart
thread_local! {
    static COOKIE: Cookie = {
        let cookie = Cookie::open(magic::CookieFlags::MIME_TYPE).unwrap();
        // Load the default database
        cookie.load::<&str>(&[]).unwrap();
        cookie
    };
}

// Setup the config globally because I can't figure out how to pass it to functions I don't call directly.
// This is evaluated at runtime, and not compilation. \o/
lazy_static! {
    pub static ref CONFIG: AppConfig = {
        let config = Figment::new()
            .merge(Toml::file("config.toml"))
            .extract()
            .unwrap_or_else(|error| {
                eprintln!("Error loading configuration: {}", error);
                std::process::exit(1);
            });

        tracing::debug!("full config: {:?}", config);
        config
    };
}

#[tokio::main]
async fn main() {
    // Read log level from config
    let log_config = &CONFIG.logging.level;

    // Construct tracing_subscriber with the filter from config.
    tracing_subscriber::registry()
        .with(fmt::layer().compact())
        .with(EnvFilter::new(log_config))
        .init();

    // Determine what SQL engine we're running.
    let sql_backend = &CONFIG.database.sql_backend;

    let db_url = &CONFIG.database.url;

    // Match on the database type
    match DatabaseType::from_str(sql_backend) {
        Some(DatabaseType::Sqlite) => {
            tracing::info!("Using SQLite backend");
            let pool = SqlitePool::connect(db_url).await.unwrap();
            SQLITE.set(pool).unwrap();
        }
        Some(DatabaseType::Postgres) => {
            tracing::info!("Using Postgres backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("Postgres not implemented yet")
            // let pool = PgPool::connect(&psql).await;
            // TODO: figure out how to use sqlx::Pool struct https://docs.rs/sqlx/latest/sqlx/struct.Pool.html
        }
        Some(DatabaseType::Mysql) => {
            tracing::info!("Using MySQL backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("MySQL not implemented yet")
        }
        Some(DatabaseType::Mssql) => {
            tracing::info!("Using MSSQL backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            unimplemented!("MSSQL not implemented yet, and honestly my hatred for microsoft will probably keep it this way")
        }
        Some(DatabaseType::Oraclecloud) => {
            tracing::info!("Using OracleCloud backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("OracleCloud not implemented yet")
        }
        _ => panic!("Unsupported SQL backend: {}", sql_backend),
    };

    // Initialise the cleaner task
    tracing::info!(
        "Starting cleaner thread every {} seconds",
        CONFIG.operations.cleaner_interval
    );
    cleaner_thread(CONFIG.operations.cleaner_interval);

    // Attempt to create the files directory
    create_dir_all("files").unwrap();

    // Internal routing
    let router = Router::new()
        // Main / functions
        .get(handlers::index::index)
        .post(handlers::upload::upload)
        // Static Pages
        .push(Router::with_path("/services").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/about").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/czb").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/qr").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/faq").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/dmca").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/welcome").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/metrics").get(handlers::serve_metrics::serve_metrics))
        // Might want to actually make a .ico instead of serving a webp image.
        .push(Router::with_path("/favicon.ico").get(StaticFile::new("static/favicon32.webp")))
        .push(Router::with_path("/robots.txt").get(StaticFile::new("static/robots.txt")))
        // Static File Serving
        .push(Router::with_path("static/<**path>").get(StaticDir::new("static/").listing(true)))
        // Deletion API
        .push(
            Router::with_path("/delete/<adminkey>")
                .delete(handlers::delete_file::delete_file)
                .get(handlers::delete_file::delete_file),
        )
        .push(
            Router::with_path("/edit/<adminkey>")
                .delete(handlers::delete_file::delete_file)
                .get(handlers::delete_file::delete_file),
        )
        // List of files the requester IP has uploaded.
        .push(Router::with_path("/my_files").get(handlers::list_files::list_files))
        // Serving uploaded files, or 404.
        .push(Router::with_path("<file>").get(handlers::serve_file::serve_file));

    // Read environment variables for host and port
    let host = &CONFIG.server.host;
    let port = CONFIG.server.port;
    let server_url = format!("{}:{}", host, port);
    tracing::info!("Listening on http://{}", server_url);
    // TODO: Allow us to listen on a socket too?
    // let acceptor = UnixListener::new("/tmp/salvo.sock").bind().await;
    let acceptor = TcpListener::new(&server_url).bind().await;

    Server::new(acceptor).serve(router).await;

    // Close SQLite before closing
    tracing::info!("Attempting to safely shut down Ephemeral.");
    SQLITE
        .get()
        .expect("Problem while safely closing the database :(")
        .close()
        .await;
}
